__version__ = "0.3.0"
__description__ = "Discrete Distribution Networks with Evolutionary Algorithms"
__license__ = "MIT OR Apache-2.0"
__author__ = "Emre Dinc"
__author_email__ = "emredinc09@gmail.com"
__maintainer__ = "Emre Dinc"
__maintainer_email__ = "emredinc09@gmail.com"
__github_username__ = "EmreDinc10"
__url__ = "https://github.com/EmreDinc10/ddn_evo"
__support__ = "https://github.com/EmreDinc10/ddn_evo/issues"
__classifiers__ = [
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
]
